#
#Licensed to the Apache Software Foundation (ASF) under one
#or more contributor license agreements.  See the NOTICE file
#distributed with this work for additional information
#regarding copyright ownership.  The ASF licenses this file
#to you under the Apache License, Version 2.0 (the
#"License"); you may not use this file except in compliance
#with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing,
#software distributed under the License is distributed on an
#"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#KIND, either express or implied.  See the License for the
#specific language governing permissions and limitations
#under the License.
#
require "logstash/outputs/base"
require "logstash/namespace"
require "logstash/environment"
require "fileutils"
require "thread"

jar_path = File.expand_path(File.join(File.dirname(__FILE__), "../../.."))
LogStash::Environment.load_runtime_jars! File.join(jar_path, "vendor")

# Datahub output plugin
class LogStash::Outputs::Datahub < LogStash::Outputs::Base
  declare_threadsafe!

  config_name "datahub"

  # datahub access id
  config :access_id, :validate => :string, :required => true

  # datahub access key
  config :access_key, :validate => :string, :required => true

  # datahub service endpoint
  config :endpoint, :validate => :string, :required => true

  # datahub project name
  config :project_name, :validate => :string, :required => true

  # datahub topic name
  config :topic_name, :validate => :string, :required => true

  # 重试次数，-1为无限重试、0为不重试、>0表示需要有限次数
  config :retry_times, :validate => :number, :required => false, :default => -1

  # 重试周期，下一次重试的间隔，单位为秒
  config :retry_interval, :validate => :number, :required => false, :default => 5

  # 由datahub异常导致的重试，超出重试次数后，是否跳过当前record
  config :skip_after_retry, :validate => :boolean, :required => false, :default => false

  # 单请求的预估bytes，达到该值就发送一个请求
  config :approximate_request_bytes, :validate => :number, :required => false, :default => 2 * 1024 * 1024

  # 按照指定字段的值计算hash，依据于该hash值落某个shard
  config :shard_keys, :validate => :array, :required => false, :default => []

  # 指定数据落指定的shard
  config :shard_id, :validate => :string, :required => false, :default => ""

  #   # 提交的列名，用户可以配置topic的列，采集部分列或者全部列
  #   # 默认为空数组，表示按照topic的顺序及全字段提交
  #   # 另外：列的配置不用保序，但是要求该字段在topic的schema中存在
  #   config :column_names, :validate => :array, :required => false, :default => []

  # 当出现脏数据时，是否继续写入
  # 当开启该开关，必须指定@dirty_data_file文件
  config :dirty_data_continue, :validate => :boolean, :required => false, :default => false

  # 脏数据文件名称，当数据文件名称，在@dirty_data_continue开启的情况下，需要指定该值
  # 特别注意：脏数据文件将被分割成两个部分.part1和.part2，part1作为更早的脏数据，part2作为更新的数据
  config :dirty_data_file, :validate => :string, :required => false

  # 脏数据文件的最大大小，该值保证脏数据文件最大大小不超过这个值，目前该值仅是一个参考值
  config :dirty_data_file_max_size, :validate => :number, :required => false, :default => 50024000

  # 数据传输压缩方式选择，目前支持deflate, lz4格式 
  config :compress_method, :validate => :string, :required => false, :default => ""

  # 是否开启pb传输，pb传输目前不支持shard_keys
  config :enable_pb, :validate => :boolean, :required => false, :default => true

  # 最大连接数
  config :max_conn, :validate => :number, :required => false, :default => 10

  # 该值内部使用，不提供配置
  # 分发shard的游标
  attr_accessor :shard_cursor

  DatahubPackage = com.aliyun.datahub.client

  public

  # def register
  def register
    begin
      # Shard cursor lock
      @shard_lock = Mutex.new

      @refresh_time = 0

      # Topic schema lock
      @schema_lock = Mutex.new

      @schema_timestamp = 0

      # 写文件锁
      @file_lock = Mutex.new

      @account = DatahubPackage.auth.AliyunAccount::new(@access_id, @access_key)
      @datahub_conf = DatahubPackage.common.DatahubConfig::new(@endpoint, @account, @enable_pb)
      @http_conf = DatahubPackage.http.HttpConfig::new()
      # @http_conf.setMaxConnTotal(@max_conn)
      # @http_conf.setMaxConnPerRoute(@max_conn)

      if @compress_method == "deflate"
        @http_conf.setCompressType(DatahubPackage.http.HttpConfig::CompressType::DEFLATE)
      elsif @compress_method == "lz4"
        @http_conf.setCompressType(DatahubPackage.http.HttpConfig::CompressType::LZ4)
      end

      @builder = DatahubPackage.DatahubClientBuilder.newBuilder()
      @builder.setDatahubConfig(@datahub_conf)
      @builder.setHttpConfig(@http_conf)

      @client = @builder.build()

      @shards = get_active_shards()
      @shard_count = @shards.size()

      @shard_cursor = rand(@shard_count)

      get_topic_schema()

      # 前置校验参数
      check_params()

      if @shard_count == 0
        @logger.error "No active shard available, please check"
        raise "No active shard available, please check"
      end

      @logger.info "Init datahub success!"
    rescue => e
      @logger.error "Init failed!" + e.message + " " + e.backtrace.inspect.to_s
      raise e
    end
  end

  def get_topic_schema()
    now = Time.now.getutc.to_i
    @schema_lock.synchronize {
      if (@schema_timestamp < now)
        @logger.info thread_id + "refresh topic schema"
        topic_meta = @client.getTopic(@project_name, @topic_name)
        @schema = topic_meta.getRecordSchema()
        fields = @schema.getFields()
        @columns_size = fields.size
        @columnnames = []
        for i in 0...@columns_size
          @columnnames.push(fields[i].getName())
        end
        @columntypes = []
        for i in 0...@columns_size
          @columntypes.push(fields[i].getType())
        end
        @logger.info "schema: " + @columnnames.to_s
        @schema_timestamp = now
      end
    }
  end

  def check_params()
    # 如果shard_id配置了，则检查该shard是否ok
    if !@shard_id.empty?
      valid = false
      for i in 0...@shards.size
        shard_entry = @shards[i]
        if shard_entry.getShardId() == @shard_id && shard_entry.getState() == DatahubPackage.model.ShardState::ACTIVE
          valid = true
        end
      end
      if (!valid)
        @logger.error "Config shard_id not exists or state not active, check your config"
        raise "Config shard_id not exists or state not active, check your config"
      end
    end

    # 检查shard_keys字段是否合法
    if @shard_keys.size > 0
      for i in 0...@shard_keys.size
        shard_key = @shard_keys[i]
        if !@schema.containsField(shard_key)
          @logger.error "Config shard_keys contains one or one more unknown field, check your config"
          raise "Config shard_keys contains one or one more unknown field, check your config"
        end
      end
    end

    # 配置了脏数据继续，必须指定脏数据文件
    if @dirty_data_continue
      if @dirty_data_file.to_s.chomp.length == 0
        raise "Dirty data file path can not be empty"
      end
    end

  end

  # 检查并设置数据到entry中
  # 如果解析数据异常，则数据落脏数据文件
  def check_and_set_data(record_data, field_type, index, event_map, column_name)
    data = event_map[column_name]
    begin
      if field_type == DatahubPackage.model.FieldType::STRING
        record_data.setField(index, data.to_s)
      elsif field_type == DatahubPackage.model.FieldType::BIGINT
        record_data.setField(index, java.lang.Long.parseLong(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::DOUBLE
        record_data.setField(index, java.lang.Double.parseDouble(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::BOOLEAN
        record_data.setField(index, java.lang.Boolean.parseBoolean(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::TIMESTAMP
        record_data.setField(index, java.lang.Long.parseLong(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::DECIMAL
        record_data.setField(index, java.math.BigDecimal::new(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::TINYINT
        record_data.setField(index, java.lang.Byte.parseByte(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::SMALLINT
        record_data.setField(index, java.lang.Short.parseShort(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::INTEGER
        record_data.setField(index, java.lang.Integer.parseInt(data.to_s))
      elsif field_type == DatahubPackage.model.FieldType::FLOAT
        record_data.setField(index, java.lang.Float.parseFloat(data.to_s))
      else
        raise "Unknown schema type of data"
      end
      return true
    rescue => e
      @logger.error "Parse column [" + column_name + "] failed, " + e.message
      # 数据格式有异常，根据配置参数确定是否续跑
      if !@dirty_data_continue
        @logger.error "Dirty data found, exit process now."
        puts "Dirty data found, exit process now."
        Process.exit(1)
        # 忽略的异常数据直接落文件
      else
        write_as_dirty_data(event_map)
      end
      return false
    end
  end

  # 脏数据文件处理
  def write_as_dirty_data(event_amp)
    dirty_file_part1_name = @dirty_data_file + ".part1"
    dirty_file_part2_name = @dirty_data_file + ".part2"

    # 加锁写入
    @file_lock.synchronize {
      dirty_file_part2 = File.open(dirty_file_part2_name, "a+")
      dirty_file_part2.puts(event_amp.to_s)
      dirty_file_part2.close
      if File.size(dirty_file_part2_name) > @dirty_data_file_max_size / 2
        # .part1, .part2分别存储数据
        # 旧数据落part1，新的数据落part2
        FileUtils.mv(dirty_file_part2_name, dirty_file_part1_name)
      end
    }
  end

  def get_active_shards()
    shard_list = @client.listShard(@project_name, @topic_name).getShards()
    active_shards = []
    for i in 0...shard_list.size
      entry = shard_list.get(i)
      if entry.getState() == DatahubPackage.model.ShardState::ACTIVE
        active_shards.push(entry)
      end
    end
    return active_shards
  end

  def get_next_shard_id()
    if !@shard_id.empty?
      return @shard_id
      # 否则轮询写入shard
    else
      idx = 0
      @shard_lock.synchronize {
        idx = @shard_cursor % @shard_count
        @shard_cursor = idx + 1
        if @shard_cursor > @shard_count
          @shard_cursor -= @shard_count
        end
      }
      shard_id = @shards[idx].getShardId()
      return shard_id
    end
  end

  def set_shard_id(entries, shard_id)
    if @shard_keys.size == 0
      entries.each {|entry| entry.setShardId(shard_id)}
    end
  end

  def send_records(entries)
    retry_count = 0

    begin
      # 提交列表必须有数据
      shard_id = get_next_shard_id()
      if entries.size > 0
        set_shard_id(entries, shard_id)
        put_result = @client.putRecords(@project_name, @topic_name, entries)
        if put_result.getFailedRecordCount() > 0
          @logger.info thread_id + "Put " + put_result.getFailedRecordCount().to_s + " records to datahub failed, total " + entries.size().to_s
          sleep @retry_interval
          entries = put_result.getFailedRecords()
          e = DatahubPackage.exception.DatahubClientException::new(
              5001, # http status
              put_result.getRequestId().to_s, # reqId
              put_result.getPutErrorEntries()[0].getErrorcode().to_s, # errCode for record 0
              put_result.getPutErrorEntries()[0].getMessage().to_s # errMessage for record 0
          )
          raise e
        end
        @logger.info thread_id + "Put data to datahub success, total " + entries.size.to_s
      end

    rescue DatahubPackage.exception.DatahubClientException => e
      @logger.error thread_id + "Flush data exception: " + e.message

      # schema 变化，需要重新获取schema
      if e.getErrorCode().to_s == "MalformedRecord" and e.getErrorMessage().to_s.include? "field size not match"
        get_topic_schema()
        resetRecordSchema(entries)
      end

      # shard的状态改变，需要重新加载shard
      if e.getErrorCode().to_s == "InvalidShardOperation"

        now = Time.now.getutc.to_i
        @shard_lock.synchronize {
          if (@refresh_time < now)
            @logger.warn thread_id + "refresh shard info"
            @shards = get_active_shards()
            @shard_count = @shards.size()
            @refresh_time = now
          end
        }

        if @shard_count == 0
          @logger.error thread_id + "No active shard available, please check"
        end
      elsif e.getErrorCode() == nil
        sleep @retry_interval
      end

      retry_count += 1
      if @retry_times >= 0 and retry_count > @retry_times and @skip_after_retry
        @logger.warn thread_id + "Retry over: " + @retry_times.to_s + ", skip entries: " + entries.size().to_s
      else
        @logger.warn thread_id + "Now retry: " + retry_count.to_s
        sleep @retry_interval
        retry
      end
    rescue => e
      @logger.error thread_id + "Flush data exception: " + e.message + " " + e.backtrace.inspect.to_s

      # 无限重试
      if @retry_times < 0
        retry_count += 1
        @logger.warn thread_id + "Now retry: " + retry_count.to_s
        # puts "Now retry..."
        sleep @retry_interval
        retry
      elsif @retry_times == 0
        @logger.error thread_id + "Retry not work, now exit"
        Process.exit(1)
        # 继续重试
      elsif @retry_times > 0
        retry_count += 1
        if retry_count > @retry_times
          @logger.warn thread_id + "Retry over: " + @retry_times.to_s
          Process.exit(1)
        end
        @logger.warn thread_id + "Now retry..."
        sleep @retry_interval
        retry
      end
    end
  end

  def resetRecordSchema(record_entries)
    record_entries.each do |entry|
      old_data = entry.getRecordData()
      old_schema = old_data.getRecordSchema()
      new_data = DatahubPackage.model.TupleRecordData::new(@schema)
      for i in 0...old_schema.getFields().size() do
        new_data.setField(i, old_data.getField(i))
      end
      entry.setRecordData(new_data)
    end
  end

  def multi_receive(event_list)
    record_entries = []
    entries_bytes = []
    invalid_record_num = 0

    event_list.each do |event|
      if event == LogStash::ShutdownSignal
        return
      end
      event_map = event.to_hash
      entry_bytes = 120

      entry = DatahubPackage.model.RecordEntry::new()
      entry.addAttribute("srcId", event_map["host"].to_s)
      entry.addAttribute("ts", event_map["@timestamp"].to_s)
      entry.addAttribute("version", event_map["@version"].to_s)
      entry.addAttribute("srcType", "log")

      record_data = DatahubPackage.model.TupleRecordData::new(@schema)
      entry.setRecordData(record_data)

      is_column_match = false
      is_data_valid = false
      for i in 0...@columns_size do
        column_name = @columnnames[i]
        column_type = @columntypes[i]
        value = event_map[column_name]
        if value != nil
          is_column_match = true
          is_data_valid = check_and_set_data(record_data, column_type, i, event_map, column_name)
          entry_bytes += value.to_s.bytesize + column_name.bytesize + 10
          break if !is_data_valid
        end
      end

      unless is_column_match
        @logger.warn thread_id + "no column match in event"
      end

      if is_data_valid
        if @shard_keys.size > 0
          hash_string = ""
          for i in 0...@shard_keys.size
            shard_key = @shard_keys[i]
            if event_map[shard_key] != nil
              hash_string += event_map[shard_key].to_s + ","
            end
          end
          hashed_value = java.lang.String.new(hash_string).hashCode()
          entry.setPartitionKey(hashed_value)
        end
        record_entries.push(entry)
        entries_bytes.push(entry_bytes)
      else
        invalid_record_num += 1
      end
    end

    if invalid_record_num > 0
      @logger.warn thread_id + "invalid record num: " + invalid_record_num.to_s
    end
    # puts "total: " + entries.size.to_s

    index = 0

    while index < record_entries.size do
      len = 1
      sum = entries_bytes[index]
      while index + len < record_entries.size do
        if sum + entries_bytes[index + len] > @approximate_request_bytes
          break
        end
        sum += entries_bytes[index + len]
        len += 1
      end

      send_records(record_entries[index, len])
      index += len
    end
  end

  # def multi_receive

  def thread_id
    "[" + Thread.current.object_id.to_s + "]"
  end

end # class LogStash::Outputs::Datahub
